from kafka import KafkaProducer
from pyspark.sql import DataFrame
from pyspark.sql.functions import col, from_json, to_json, struct
from typing_extensions import deprecated

from mx_stream_core.config.kafka import kafka_bootstrap_servers
from mx_stream_core.config.app import service_name

# Create Producer instance
_producer = None


def get_kafka_producer():
    """
    Get a Kafka producer
    :return:
    """
    global _producer
    if _producer is None:
        _producer = KafkaProducer(
            bootstrap_servers=kafka_bootstrap_servers,
            value_serializer=lambda v: str(v).encode('utf-8')
        )
    return _producer


def read_stream_from_kafka(spark, topic, schema) -> DataFrame:
    """
    Read a stream from Kafka
    :param spark: SparkSession
    :param topic: Kafka topic
    :param schema: Schema of the data
    :return:
    """
    df = spark \
        .readStream \
        .format("kafka") \
        .option("kafka.bootstrap.servers", kafka_bootstrap_servers) \
        .option("subscribe", topic) \
        .option("startingOffsets", "earliest") \
        .load()
    df = df.selectExpr("CAST(value AS STRING) as json_str")
    df = df.withColumn("json_data", from_json(col("json_str"), schema)).select("json_data.*")
    return df


def write_stream_to_kafka(df, topic, checkpoint_path):
    """
    Write a stream to Kafka
    :param df: DataFrame
    :param topic: Kafka topic
    :param checkpoint_path: Checkpoint path for the stream
    :return:
    """
    # Convert DataFrame to JSON strings
    json = df.select(to_json(struct(
        *df.columns
    )).alias("value"))

    # Write the merged stream to Kafka
    query = json \
        .writeStream \
        .format("kafka") \
        .option("kafka.bootstrap.servers", kafka_bootstrap_servers) \
        .option("topic", topic) \
        .option("checkpointLocation", checkpoint_path) \
        .start()

    return query


def delivery_report(err, msg):
    """
    Reports the result of a message delivery attempt.
    :param err: The error that occurred or None if the message was delivered successfully.
    :param msg: The message that was sent or failed to send.
    :return:
    """
    if err is not None:
        print('Message delivery failed: {}'.format(err))
    else:
        print('Message delivered to {} [{}]'.format(msg.topic(), msg.partition()))


def produce_kafka_message(topic, message):
    """
    Produce a message to Kafka
    :param topic: Kafka topic
    :param message:  Message to send
    :return:
    """
    producer = get_kafka_producer()
    producer.send(topic, value=message).add_errback(delivery_report)
    producer.flush()  # Wait for all messages to be delivered


def produce_kafka_messages(topic, event_name: str, df: DataFrame):
    """
    Produce a message to Kafka
    :param topic: Kafka topic
    :param message:  Message to send
    :return:
    """
    producer = get_kafka_producer()
    for row in df.toLocalIterator():
        message = {
            "event": event_name,
            "data": row.asDict()
        }
        producer.send(topic, value=message).add_errback(delivery_report)
    producer.flush()  # Wait for all messages to be delivered


@deprecated("Use create_ingestor_topic_name or create_transformation_topic_name instead")
def create_topic_name(ingestor_name: str) -> str:
    """
    Create a topic name
    :param ingestor_name: Name of the ingestor (e.g. "word")
    :return:
    """
    return f'{service_name}_ingestor_{ingestor_name}'


def create_ingestor_topic_name(ingestor_name: str) -> str:
    """
    Create a topic name for an ingestor
    :param ingestor_name: Name of the ingestor (e.g. "word")
    :return:
    """
    return f'{service_name}_ingestor_{ingestor_name}'


def create_transformation_topic_name(transformation_name: str) -> str:
    """
    Create a topic name for a transformation
    :param transformation_name: Name of the transformation (e.g. "book")
    :return:
    """
    return f'{service_name}_transformation_{transformation_name}'
